/// <reference path="../../typings/index.d.ts" />

import * as express from 'express';
import * as bodyParser from 'body-parser';
import * as logger from 'morgan';
import * as cookieParser from 'cookie-parser';
import * as http from 'http';

import {
    Request,
    Response
} from 'express';

export class Application {
    private static instance: Application;
    private express: any;

    public Host: string = 'localhost';
    public Port: number = 3000;
    public Prefix: string;

    constructor() {
        Application.instance = this;

        this.initializeExpress();
        this.onBeforeApplicationStart();
    }

    // -- Initliaze

    initializeExpress() {
        let app: any = express();

        app.use(logger('dev'));
        app.use(bodyParser.json());
        app.use(bodyParser.urlencoded({ extended: false }));
        app.use(cookieParser());

        app.get('/', (req: Request, res: any) => {
            res.send('it worked');
        });

        this.express = app;
    }

    initializeErrorHandler() {
        this.express.use('*', (req: any, res: any, next: any) => {
            res.status(404)
                .json({
                    status: 404,
                    message: 'Not Found',
                    stacktrace: new Error().stack
                });
        });

        this.express.use((err: any, req: any, res: any, next: any) => {
            if (err) {
                console.error(err.stack);
            }

            res.status(500)
                .json({
                    status: 500,
                    message: err.message,
                    stacktrace: err.stack
                });
        });
    }

    // -- Hooks
    onBeforeApplicationStart() {

    }

    // -- Register
    private registerGetRoutes(ResourceClass: any) {

        let getMethods: Array<any> = ResourceClass.constructor.__get;

        if (typeof getMethods !== 'undefined') {
            getMethods.forEach((getMethod) => {

                this.express.get(this.getPath(ResourceClass, getMethod), (req: any, res: any) => {
                    console.log(getMethod);
                    let resource = new ResourceClass.constructor();
                    resource.req = req;
                    resource.res = res;

                    let paramList: Array<any> = this.getParamList(ResourceClass, getMethod.methodName, req);

                    Promise.resolve(resource[getMethod.methodName].apply(resource, paramList)).then((returnValue) => {
                        res.json(returnValue);
                    }).catch((err) => res.status(err.statusCode || 500).json(err));
                });
            });
        }
    }

    private registerPostRoutes(ResourceClass: any) {
        let postMethods: Array<any> = ResourceClass.__post;

        if (typeof postMethods !== 'undefined') {

            postMethods.forEach((postMethod) => {
                this.express.post(this.getPath(ResourceClass, postMethod), (req: any, res: any) => {
                    let resource = new ResourceClass();
                    resource.req = req;
                    resource.res = res;

                    let paramList: Array<any> = this.getParamList(ResourceClass, postMethod.methodName, req);

                    Promise.resolve(resource[postMethod.methodName].apply(resource, paramList)).then((returnValue) => {
                        res.json(returnValue);
                    }).catch((err) => res.status(err.statusCode || 500).json(err));
                });
            });
        }
    }

    private registerPutRoutes(ResourceClass: any) {
        let putMethods: Array<any> = ResourceClass.__put;

        if (typeof putMethods !== 'undefined') {

            putMethods.forEach((putMethod) => {
                this.express.put(this.getPath(ResourceClass, putMethod), (req: any, res: any) => {
                    let resource = new ResourceClass();
                    resource.req = req;
                    resource.res = res;

                    let paramList: Array<any> = this.getParamList(ResourceClass, putMethod.methodName, req);

                    Promise.resolve(resource[putMethod.methodName].apply(resource, paramList)).then((returnValue) => {
                        res.json(returnValue);
                    }).catch((err) => res.status(err.statusCode || 500).json(err));
                });
            });
        }
    }

    private registerDeleteRoutes(ResourceClass: any) {
        let deleteMethods: Array<any> = ResourceClass.__delete;

        if (typeof deleteMethods !== 'undefined') {

            deleteMethods.forEach((deleteMethod) => {
                this.express.delete(this.getPath(ResourceClass, deleteMethod), (req: any, res: any) => {
                    let resource = new ResourceClass();
                    resource.req = req;
                    resource.res = res;

                    let paramList: Array<any> = this.getParamList(ResourceClass, deleteMethod.methodName, req);

                    Promise.resolve(resource[deleteMethod.methodName].apply(resource, paramList)).then((returnValue) => {
                        res.json(returnValue);
                    }).catch((err) => res.status(err.statusCode || 500).json(err));
                });
            });
        }
    }

    // -- Helpers
    private constructCorrectType(req: express.Request, paramDef: any) {
        if (typeof paramDef.constructor !== 'undefined') {
            if (paramDef.isArray === true) {
                return req.body.map((item: any) => {
                    return paramDef.constructor.fromJson(item);
                });
            } else {
                return paramDef.constructor.fromJson(req.body);
            }
        } else {
            return req.body;
        }
    }

    private getParamList(ResourceClass: any, methodName: string, req: express.Request): Array<any> {
        if (typeof ResourceClass.constructor.prototype[methodName].__params !== 'undefined') {

            return ResourceClass.constructor.prototype[methodName].__params.map((paramDef: any) => {

                if (paramDef.paramType === 'PATH_PARAM') {
                    return req.params[paramDef.paramName];
                } else if (paramDef.paramType === 'QUERY_PARAM') {
                    return req.query[paramDef.paramName];
                } else if (paramDef.paramType === 'BODY_PARAM') {

                    return this.constructCorrectType(req, paramDef);
                }
            });
        } else {
            return [];
        }
    }

    private getPath(ResourceClass: any, getMethod: any): string {
        let path: any = [];

        if (this.Prefix != null) {
            path.push(this.Prefix);
        }

        path.push(ResourceClass.constructor.__path);

        let methodPath = ResourceClass.constructor.prototype[getMethod.methodName].__path;

        if (typeof methodPath !== 'undefined') {
            path.push(methodPath);
        }

        return path.join('');
    }

    // -- Public
    public registerResource(ResourceClass: any) {
        this.registerGetRoutes(ResourceClass);
        this.registerPostRoutes(ResourceClass);
        this.registerPutRoutes(ResourceClass);
        this.registerDeleteRoutes(ResourceClass);
    }

    public start() {
        let server: http.Server = http.createServer(<any>this.express);

        server.listen(this.Port, this.Host, () => {
            console.log('server listening at ', server.address().address + ':' + server.address().port);
        });

    }
}